# -*- coding: utf-8 -*-

import pygame
from pygame.rect import Rect
from pygame.locals import *
from weakref import WeakKeyDictionary

import gui, model
from game import Game, NoFunds, NoGood
from scene import Scene
from render import MapRenderer
from sounds import play_sample
from data import get_path

class Selection (object):
	def __init__(self, rect):
		self.rect = Rect(rect)
		self.base = self.rect.topleft
		self.area = self.rect.w * self.rect.h

	def extend(self, pos):
		if pos[0] < self.base[0]:
			self.rect.left = pos[0]
			self.rect.w = self.base[0] - pos[0] + 1
		else:
			self.rect.left = self.base[0]
			self.rect.w = pos[0] - self.base[0] + 1

		if pos[1] < self.base[1]:
			self.rect.top = pos[1]
			self.rect.h = self.base[1] - pos[1] + 1
		else:
			self.rect.top = self.base[1]
			self.rect.h = pos[1] - self.base[1] + 1

		self.area = self.rect.w * self.rect.h

class SelectionDialog (gui.Group):
	def __init__(self, scene, dispatcher):
		gui.Group.__init__(self, dispatcher)
		self.selection = None
		self.scene = scene
		self.game = scene.game
		self.map = scene.game.map
		self.nogood = 0
		self.order_slots = range(20)
	
		self.l_area = gui.Label((710, 20, 130, 18), u'Selected %.2f km²')
		self.add_widget(self.l_area)

		self.b_buy = gui.Button((710, 50, 130, 20), 'Buy land')
		self.b_buy.clicked = self.make_button_callback(None)
		self.l_buy = gui.Label((850, 50, 130, 20), u'%d £')
		self.add_widgets(self.b_buy, self.l_buy)

		offset = 74
		self.buttons = {}
		for t in self.map.terraininfo.types.values():
			if not t.buildable:
				continue

			b = gui.Button((710, offset, 130, 20), t.button)
			b.clicked = self.make_button_callback(t)

			c = gui.Label((850, offset, 130, 20), u'%d £')
			self.buttons[t.tid] = (b, c)
			self.add_widgets(b, c)

			offset += 24


		self.disable()

	def make_button_callback(self, target):
		# Closures are straight way to insanity.
		def callback(btn, pos):
			if btn != 1:
				return

			# We'll do whith this another time.
			if not self.order_slots:
				self.scene.add_message(64, "Too many orders!", (255, 255, 0))
				return

			selection = self.selection.rect

			try:
				if target is not None:
					order = self.game.player.create_order(selection, target)
				else:
					self.game.player.buy_land(selection)
			except NoFunds:
				self.scene.add_message(64, "Insufficient funds!", (255, 255, 0))
				return
			except NoGood:
				self.scene.add_message(64, "City is in the way!", (255, 255, 0))
				return

			self.scene.deselect()

			if target is None:
				return

			slot = self.order_slots.pop(0)
			y = 50 + 24 * slot 
			
			text = u'%.2f km² to %s' % (order.area * 0.01, target.name) 
			button = gui.ProgressButton((10, y, 200, 20), text, 'CANCEL')
			self.dispatcher.add_widget(button)

			def free_slot():
				self.order_slots.append(slot)
				self.order_slots.sort()
				self.dispatcher.del_widget(button)

			order.on_complete = free_slot

			def on_draw():
				button.update(float(order.done) / order.area)
			button.on_draw = on_draw

			def cancel_order(btn, pos):
				if btn != 1:
					return
				order.cancel()
				# Told ya.
				free_slot()

			button.clicked = cancel_order

		return callback

	def set_selection(self, selection):
		self.selection = selection
		if selection:
			self.enable()
			self.l_area.update(selection.area * 0.01)

			rinfo, land_cost, nogood = self.map.get_rect_info(selection.rect)
			self.l_buy.update(land_cost)
			self.nogood = nogood

			for tid, (button, label) in self.buttons.items():
				cost = self.map.get_project(tid, rinfo)
				label.update(cost)

		else:
			self.disable()
			self.nogood = 0
	
class Message (object):
	def __init__(self, size, text, color, time=0, font=None):
		self.surf = gui.render_text(font, size, text, color)
		self.x = 512 - self.surf.get_width() / 2
		self.y = 768
		self.h = self.surf.get_height()
		self.speed = 20
		self.halt = (self.y - self.h) / 2
		self.time = time
		self.gone = False
	
	def think(self):
		if self.y > self.halt:
			self.y = int(4. * self.y + self.halt) / 5.
		else:
			if self.time:
				self.time -= 1
			else:
				self.y -= self.speed
				self.speed += 20

			if self.y < -self.h:
				self.gone = True

	def draw(self, surf):
		surf.blit(self.surf, (self.x, self.y))
		
class PlayScene (Scene):
	"""Scene with actual gameplay.
	"""
	big_font = 'DejaVuSans-Bold.ttf'

	def init(self):
		Scene.init(self)

		self.game = Game()
		self.map = self.game.map
		self.game.player.got_cash = self.got_cash
		self.game.computer.started_over = self.started_over
		
		self.renderer = MapRenderer(self.map)
		self.renderer.load_sprites()

		pw, ph = self.renderer.get_prerender_size()
		self.sw, self.sh = sw, sh = self.manager.display.get_size()
	
		self.last_cash = self.game.player.cash

		self.buf = pygame.Surface((pw, ph))
		self.renderer.prerender(self.buf)

		self.lx = pw - sw - 1
		self.ly = ph - sh - 1

		# View and target positions.
		self.vx = self.tx = 0
		self.vy = self.ty = 0

		# Dispatcher for GUI.
		self.dispatcher = gui.Dispatcher()
		self.dispatcher.default = self

		# Selection dialog
		self.selection = None
		self.dialog = SelectionDialog(self, self.dispatcher)

		self.messages = []
		self.add_message(72, 'GO!', (255, 0, 0), font=self.big_font)

		self.dozers = WeakKeyDictionary()
		self.dozer_sprites = {}

		for face in ['n', 's', 'e', 'w']:
			self.dozer_sprites[face] = pygame.image.load(get_path('dozer-%s.png' % face)).convert_alpha()

		self.over = None
	
	def add_message(self, size, text, color, time=0, font=None):
		self.messages.append(Message(size, text, color, time, font))

	def got_cash(self, cash):
		self.add_message(32, u'%d £' % cash, (255, 255, 255))

	def started_over(self):
		self.add_message(32, 'City emergency expansion plan activated.', (255, 255, 255))

	def clicked(self, button, pos):
		if button == 1:
			self.selection = None
			self.dialog.set_selection(None)
	
	def focus(self):
		pass
	
	def blur(self):
		pass

	def motion(self, buttons, pos, rel):
		pass

	def dragging(self, button, start, pos, rel):
		if button == 1:
			start = self.unproject(start)
			mx, my = self.unproject(pos)
			if not self.selection or start != self.selection.base:
				self.deselect()
				self.selection = Selection((mx, my, 1, 1))
			else:
				self.selection.extend((mx, my))

		if button == 3:
			self.tx -= rel[0]
			self.ty -= rel[1]

	def dragged(self, button, start, pos):
		if button == 1:
			self.dialog.set_selection(self.selection)

	def deselect(self):
		self.selection = None
		self.dialog.set_selection(None)

	def update_view(self):
		if self.tx == self.vx and self.ty == self.vy:
			return

		if self.tx < 0:
			self.tx = 0
		if self.ty < 0:
			self.ty = 0

		if self.tx > self.lx:
			self.tx = self.lx
		if self.ty > self.ly:
			self.ty = self.ly

		if -4 < self.vx - self.tx < 4:
			self.vx = self.tx
		else:
			self.vx = int((4. * self.vx + self.tx) / 5. + .5)

		if -4 < self.vy - self.ty < 4:
			self.vy = self.ty
		else:
			self.vy = int((4. * self.vy + self.ty) / 5. + .5)

	def project(self, pos):
		return (pos[0] * self.renderer.spr_w - self.vx, pos[1] * self.renderer.spr_h - self.vy)

	def projectrect(self, rect):
		w, h = self.renderer.spr_w, self.renderer.spr_h
		return (rect[0] * w - self.vx, rect[1] * h - self.vy, rect[2] * w, rect[3] * h)

	def unproject(self, pos):
		return (pos[0] + self.vx) / self.renderer.spr_w, (pos[1] + self.vy) / self.renderer.spr_h

	def update_cash(self):
		cash = self.game.player.cash
		if -10 < self.last_cash - cash < 10:
			self.last_cash = cash
		else:
			self.last_cash = int((6. * self.last_cash + cash) / 7.)

	def think(self):
		for ev in pygame.event.get():
			if ev.type == KEYDOWN and ev.key == K_F12:
				self.game.player.payday()
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				return 'pause', (self,)

			self.dispatcher.handle_event(ev)

		self.messages = [m for m in self.messages if not m.gone]
		self.game.think()
		self.update_cash()
		self.update_view()

		if self.over is not None:
			self.over -= 1
			if self.over == 0:
				return self.game.over, (self,)

		elif self.game.over:
			self.game_over()
			self.over = 150

	def game_over(self):
		why = self.game.over
		if why == 'won':
			self.add_message(72, 'You won!', (0, 64, 0), 100, self.big_font)
		else:
			self.add_message(72, 'Game over!', (0, 64, 0), 100, self.big_font)
	
	def render(self, surf):
		# Update map buffer.
		self.renderer.update(self.buf)

		# Draw map.
		surf.blit(self.buf, (0, 0), (self.vx, self.vy, surf.get_width(), surf.get_height()))

		# Show dozers and orders.
		for o in self.game.player.orders:
			for d in o.dozers:
				if not d.position:
					continue

				x, y = pos = d.position
				face = 'e'
				if d in self.dozers:
					ppos, pface = self.dozers[d]
					if pos == ppos:
						face = pface
					else:
						px, py = ppos
						if x < px:
							face = 'w'
						elif y < py:
							face = 'n'
						elif y > py:
							face = 's'
					
				pd = self.project(pos)
				surf.blit(self.dozer_sprites[face], pd)
				self.dozers[d] = (pos, face)

			r = self.projectrect(o.rect)
			surf.set_clip()
			pygame.draw.rect(surf, (255, 255, 0), r, 2)

		# Render a selection.
		if self.selection:
			if self.dialog.nogood:
				color = (255, 0, 0)
			else:
				color = (255, 255, 255)
			r = self.selection.rect
			pygame.draw.rect(surf, color, self.projectrect(r), 2)
		
		text = gui.render_text(None, 20, u'%d £' % self.last_cash, (255, 255, 255))
		surf.blit(text, (10, 10))
	
		text = gui.render_text(None, 20, u'Fame: %d' % self.game.player.fame, (255, 255, 255))
		surf.blit(text, (200, 10))

		text = gui.render_text(None, 18, str(self.game.date), (255, 255, 255))
		surf.blit(text, (1000 - text.get_width(), 745))

		# GUI overlay.
		self.dispatcher.draw_all(surf)

		# Messages.
		for m in self.messages:
			m.think()
			m.draw(surf)

